# -*- coding: utf-8 -*-

"""Implementation of the Missionaries and Cannivals problem."""

from hlogedu.search.problem import action, DDInterval, Problem


class MissionariesAndCannibalsProblem(Problem):
    """Missionaries and Cannibals Problem

    In this problem we have a given number of missionaries and
    the same number of cannibals on one side of a river.

    The goal is to transport all the individuals to the other side.
    To do so, there is a boat that can fit two individuals, but at
    any given time there cannot be more cannibals than missionaries
    on any side. Also, someone has to row the boat, thus the boat
    can never traverse the river empty.

    The state is a tuple of three elements:
        - # of missionaries on the wrong side
        - # of cannibals on the wrong side
        - The boat is in the wrong side True/False
    """

    NAME = "MCP"

    def __init__(self):
        self.n_individuals = 3
        self.boat_size = 2

    def get_start_states(self):
        """Creates the initial state.

        The state is defined as a tuple of two integers and a boolean. The
        first integer is the number of missionaries in the wrong side, the
        second integer the number of cannibals on the wrong side and the
        boolean whether the boat is in the wrong side.
        """
        return [(self.n_individuals, self.n_individuals, True)]

    def is_goal_state(self, state):
        """Tests if the state is a goal state.

        To be a goal state, all the missionaries and cannibals must be on the
        right side of the river.
        """
        return state[0] == 0 and state[1] == 0

    def is_valid_state(self, state):
        """Verifies that the given state is valid.

        In this problem a state is valid if the cannibals do not eat the
        missionaries in both sides of the river.
        """
        valid_range = 0 <= state[0] <= self.n_individuals \
            and 0 <= state[1] <= self.n_individuals

        # Cannibals do not eat missionaries on any shore
        left_side_ok = state[0] == 0 or state[0] >= state[1]
        right_side_ok = state[0] == self.n_individuals or state[0] <= state[1]

        return valid_range and left_side_ok and right_side_ok

    @action(DDInterval(0, "boat_size"), DDInterval(0, "boat_size"), cost=1)
    def cross(self, state, missionaries, cannibals):
        """Makes the boat cross the river.

        Any time the boat crosses the river it must carry at least one
        individual and never more than 'boat_size'.
        """
        m_on_shore = state[0] if state[2] else self.n_individuals - state[0]
        c_on_shore = state[1] if state[2] else self.n_individuals - state[1]
        n_boaters = missionaries + cannibals

        if missionaries > m_on_shore or cannibals > c_on_shore or \
                n_boaters <= 0 or n_boaters > self.boat_size:
            return None

        if state[2]:
            return state[0] - missionaries, state[1] - cannibals, False
        return state[0] + missionaries, state[1] + cannibals, True
