# -*- coding: utf-8 -*-


"""Different implementations of the N-Queens problem."""

from random import randint

# Modules/Packages provided by the tool.
from hlogedu.search.problem import action, DRange, Heuristic, Problem


class NQueensGreedy(Problem):
    """N-Queens problem

    This problem consists in placing n non-attacking queens on an
    nxn chessboard.

    This implementations starts with an emtpy nxn chessboard, and tries to
    solve the problem by placing one queen at a time into any of the empty
    possitions.
    """

    NAME = "NQueensG"

    def __init__(self):
        super().__init__()
        self.n_queens = 8
        self.b_size = max(8, self.n_queens)

    def get_start_states(self):
        """Creates and returns the initial state.

        The state is representet as a tuple of integers, each integer
        represents a row and the index of the integer a column. If
        the 3rd position of the tuple contains a 2 it means that a queen
        is located into the 3rd column at the 2nd row.

        Negative values mean that there is no queen in that column.
        """
        return [(-1,) * self.b_size]

    def is_goal_state(self, state):
        """Tests if state is a goal state.

        Assuming that all states given to this function are valid
        we just need to check that all positions are non-negative.
        """
        return all(x >= 0 for x in state)  # Not correct update

    def is_valid_state(self, state):
        """Tests whether or not all queens are in non-attacking state."""
        for i in range(len(state)):
            if state[i] < 0:  # Ignore column if it hasn't a queen
                continue

            for j in range(i + 1, len(state)):
                ri, rj = state[i], state[j]
                diagonal_check = abs(i - j) == abs(ri - rj)
                if rj > -1 and (ri == rj or diagonal_check):
                    return False

        return True

    @action(DRange(0, 8), DRange(0, 8), cost=1)
    def place(self, state, row, column):
        """Places a queen in the specified position."""
        if state[column] >= 0:
            return None

        return tuple(r if i != column else row for i, r in enumerate(state))


class NQueensIterativeRepair(Problem):
    """N-Queens problem

    This problem consists in placing n non-attacking queens on an
    nxn chessboard.

    This implementations starts with an nxn chessboard that already
    contains N queens on it, and tries to solve the problem by iteratively
    moving the queens to different possitions
    """

    NAME = "NQueensIR"

    def __init__(self):
        super().__init__()
        self.n_queens = 8
        self.b_size = max(8, self.n_queens)

    def get_start_states(self):
        return [
            tuple(randint(0, self.b_size - 1) for _ in range(self.b_size))
        ]

    def is_goal_state(self, state):
        count_conflicts(state) == 0

    def is_valid_state(self, state):
        return all(0 <= pos <= self.b_size for pos in state)

    @action(DRange(0, 7), DRange(0, 7), cost=1)
    def move(self, state, row, column):
        if state[column] == row:  # Not moving
            return None

        return tuple(r if i != column else row for i, r in enumerate(state))


@NQueensIterativeRepair.heuristic
class RepairHeuristic(Heuristic):

    def compute(self, state):
        return count_conflicts(state)


def count_conflicts(state):
    n_conflicts = 0

    for i in range(len(state)):
        if state[i] < 0:  # Ignore column if it hasn't a queen
            continue

        for j in range(i + 1, len(state)):
            ri, rj = state[i], state[j]
            diagonal_check = abs(i - j) == abs(ri - rj)
            if rj > -1:
                n_conflicts += 1 if ri == rj else 0
                n_conflicts += 1 if diagonal_check else 0

    return n_conflicts
