# -*- coding: utf-8 -*-

"""Implementation of the search problems from the UdL AI exam.
Course: 2015-16
"""


# Modules/Packages provided by the tool.
from hlogedu.search.containers import HashableList
from hlogedu.search.problem import Categorical, Heuristic, Problem, action


# Problem 4
##############################################################################

class AIExam1516P4(Problem):
    """Problem 4 [University of Lleida AI Exam 2015-16]
    Considera el siguiente grafo dirigido del espacio de estados de un problema
    donde el estado A es el estado inicial y G1, G2 son dos estados finales.
    Indica para los siguientes algoritmos los 5 primeros nodos expandidos en
    orden.

        (0.5 puntos) DFS, en grafo, y test de solución en sucesores.
        (0.5 puntos) BFS, en árbol, y test de solución en sucesores.
        (0.5 puntos) Búsqueda “primero el mejor” voraz, en árbol.
        (0.75 puntos) A∗ , en árbol.
        (0.25 puntos) ¿Es la heurística admisible?

    Notas:
        Los sucesores se generan en orden alfabético.
        Los empates se rompen alfabéticamente.
    """

    def __init__(self):
        self.graph = {
            ('A', 'B'): 4, ('A', 'C'): 2,
            ('B', 'D'): 2,
            ('C', 'A'): 0.5, ('C', 'E'): 3,
            ('D', 'G1'): 4, ('D', 'G2'): 10,
            ('E', 'G1'): 1, ('E', 'G2'): 10
        }

        self.heuristic = {'A': 6, 'B': 5, 'C': 2, 'D': 4, 'E': 5,
                          'G1': 0, 'G2': 0}

        self.start_state = 'A'
        self.goal_states = frozenset(['G1', 'G2'])

    def get_start_states(self):
        """Creates and returns the initial state."""
        return [self.start_state]

    def is_goal_state(self, state):
        """Tests if state is one of the goal states."""
        return state in self.goal_states

    def is_valid_state(self, _):
        """Tests if the state is valid.
        Always evaluates to true because the actions only return valid states.
        """
        return True

    @action(Categorical(['A', 'B', 'C', 'D', 'E', 'G1', 'G2']))
    def move(self, state, n_node):
        """Moves the state to another node.

        The move is valid iff there is an edge connecting the current
        and the next node.
        """
        edge = (state, n_node)

        if edge in self.graph:
            return self.graph[edge], n_node

        return None


@AIExam1516P4.heuristic
class Fixed(Heuristic):
    """Uses a fixed heuristic defined in the problem.

    The instance of the problem class can be accessed via 'self' as
    'self.problem'.
    """

    def compute(self, state):
        """Returns the heuristic value defined in the problem."""
        return self.problem.heuristic[state]


# Problem 5
##############################################################################

class AIExam1516P5(Problem):
    """Problem 5 [University of Lleida AI Exam 2015-16]
    Capacitated Vehicle Routing Problem with Pickup and Delivery (CVRPPD):
    disponemos de un camión con una capacidad de carga de k artículos, un
    conjunto de localizaciones (puntos de carga o descarga y punto de salida),
    y conexiones posibles entre las localizaciones. En cada una de estas
    localizaciones el camión debe o bien cargar o bien descargar un cierto
    tipo distinto de articulo. Dada una carga inicial y un punto de salida,
    el objetivo es determinar la ruta óptima, aquella en la que se recorren
    menos kilómetros, que tiene que realizar el camión para repartir todos los
    artículos.
    """

    def __init__(self):
        self.graph = {
            ('A', 'B'): 3, ('A', 'D'): 4, ('A', 'I'): 4,
            ('B', 'A'): 3, ('B', 'C'): 5, ('B', 'D'): 5,
            ('C', 'B'): 5, ('C', 'D'): 9, ('C', 'F'): 7,
            ('D', 'A'): 4, ('D', 'B'): 5, ('D', 'C'): 9,
            ('D', 'E'): 5, ('D', 'I'): 4,
            ('E', 'D'): 5, ('E', 'F'): 5,
            ('F', 'C'): 7, ('F', 'E'): 5,
            ('I', 'A'): 4, ('I', 'D'): 4
        }

        self.pickup_points = {'A': 1, 'B': 2, 'C': 3}
        self.delivery_points = {'D': 1, 'E': 2, 'F': 3}

        self.capacity = 2

    def get_start_states(self):
        """Creates and returns the initial state."""
        # This is extremely inefficient, but it should suffice for
        # demonstration purposes
        start_state = HashableList(["I", [], []])
        return [start_state]

    def is_goal_state(self, state):
        """Tests if all the items have been delivered."""
        for c in self.delivery_points.values():
            if c not in state[2]:
                return False
        return True

    def is_valid_state(self, _):
        """Tests if the state is valid."""
        return True

    @action(Categorical(['A', 'B', 'C', 'D', 'E', 'F', 'I']))
    def move(self, state, n_node):
        """Moves the vehicle to another node.

        The move is valid iff there is an edge connecting the current
        and the next node.
        """
        node, _, _ = state
        edge = (node, n_node)

        if edge in self.graph:
            new_state = state.copy()
            new_state[0] = n_node  # Change node
            return self.graph[edge], new_state

        return None

    @action(cost=0)
    def load(self, state):
        """Loads an item into the truck.

        In order to load an item, the truck must be in a node that provides
        any type of item, have room for it. Additionally, we don't want to
        load an item that is already on the truck or that has been delivered.

        :return: A new state in which the appropiate item has been loaded into
                 the truck.
        """
        if len(state[1]) >= self.capacity:      # Is it full?
            return None
        if state[0] not in self.pickup_points:  # Is a valid pickup point?
            return None

        # Item already loaded/delivered?
        item = self.pickup_points[state[0]]
        if item in state[1] or item in state[2]:
            return None

        new_state = state.copy()
        new_state[1].append(item)
        return new_state

    @action(cost=0)
    def deliver(self, state):
        """Delivers the item requested by the current node.

        In order to deliver an item, the node must be a deliver node and the
        item requested by the node must be into the truck.

        :return: A new state in which the appropiate item has been delivered
                 and, thus, removed from the truck.
        """
        if state[0] not in self.delivery_points:
            return None

        item = self.delivery_points[state[0]]
        if item not in state[1] or item in state[2]:
            return None

        new_state = state.copy()
        new_state[1].remove(item)
        new_state[2].append(item)
        return new_state
