# -*- coding: utf-8 -*-

"""Implementation of the search problems from the UdL AI exam.
Course: 2016-17
"""

import random
import sys

# Modules/Packages provided by the tool
from hlogedu.search.containers import HashableList
from hlogedu.search.problem import DCategorical, DDRange, Heuristic, \
                                   Problem, action


class AIExam1617P2(Problem):
    """Problem 2 [University of Lleida AI Exam 2016-17]"""

    def __init__(self):
        self.graph = {
            ('A', 'B'): 6, ('A', 'C'): 4, ('A', 'D'): 6,
            ('B', 'A'): 6, ('B', 'C'): 3, ('B', 'E'): 3,
            ('C', 'B'): 3, ('C', 'F'): 4,
            ('D', 'C'): 3, ('D', 'F'): 2, ('D', 'G1'): 15,
            ('E', 'C'): 2, ('E', 'G2'): 1,
            ('F', 'H'): 4, ('F', 'G3'): 6,
            ('H', 'G3'): 2,
        }

        self.heuristic = {
            'A': 8, 'B': 3, 'C': 4, 'D': 7, 'E': 1, 'F': 6, 'H': 2,
            'G1': 0, 'G2': 0, 'G3': 0
        }

        self.nodes = tuple(self.heuristic.keys())  # used in action
        self.start_state = 'A'
        self.goal_states = frozenset(['G1', 'G2', 'G3'])

    def get_start_states(self):
        """Creates and returns the initial state."""
        return [self.start_state]

    def is_goal_state(self, state):
        """Tests if state is one of the goal states."""
        return state in self.goal_states

    def is_valid_state(self, _):
        """Tests if the state is valid.
        Always evaluates to true because the actions only return valid states.
        """
        return True

    @action(DCategorical('nodes'))
    def move(self, state, next_node):
        """Moves the state to another node.

        The move is valid iff there is an edge connecting the current
        and the next node.
        """
        edge = (state, next_node)

        if edge in self.graph:
            return self.graph[edge], next_node

        return None


class AIExam1617P3(Problem):
    """Problem 3 [University of Lleida AI Exam 2016-17]
    Coloración de mapas: disponemos de un mapa de países que deseamos
    colorear. Dado un mapa inicial, el objetivo del problema es determinar el
    número mínimo de colores necesarios para colorear el mapa sin que dos
    países fronterizos tengan el mismo color. Define adecuadamente el problema
    anterior, indicando, el estado inicial, el test de objetivo, las acciones,
    el modelo de transición y el coste de un paso. Describe una heurística
    admisible y qué algoritmo utilizarías para resolver el problema.
    """

    def __init__(self):
        self.n_nodes = 0  # don't change the variable name (used in action)
        self.adjacency = None

        self._init_random(5, 542503521)  # size and seed
        # self._init_petersen()
        self.n_colors = self.n_nodes     # limits colors to try

        # print the adjacency matrix for debugging purposes
        import pprint
        pprint.pprint(self.adjacency, stream=sys.stderr)

    def get_start_states(self):
        """Creates and initializes the start state.

        We represent the state of this problem as a list, where each position
        has an integer that represents the color of the node. Colors are
        stored as positibe integers and -1 has the special meaning "no color".
        """
        state = HashableList([-1] * self.n_nodes)
        return [state]

    def is_goal_state(self, state):
        """Tests if all the nodes have been painted

        In order for this test to work properly, the actions must be
        sound (don't paint two adjacent nodes with the same color).
        """
        return all(c >= 0 for c in state)

    def is_valid_state(self, _):
        """Unnecessary, the actions perform all the necessary checks."""
        return True

    @action(DDRange('n_nodes'), DDRange('n_colors'))
    def paint(self, state, node, color):
        """Paints the node if it hasn't been painted yet or if the color is
        being used in any of the neighboring nodes.

        The cost of the action depends on the color being used. In particular
        to guarantee that solution is optimum, it uses the minimum number of
        colors, when using UCS/A* we define it as:

            > cost(c) = cost(c-1) * N + 1
            > cost(0) = 1

        Where c is the color being used to paint the node, which can take
        values in range [0, inf) and N the total number of nodes.

        Since Python is not recursion-friendy we use the iterative version:

            > cost(c) = sum from i=0 to c of N^i
        """
        if state[node] >= 0:  # already painted
            return None

        for neighbor in self.adjacency[node]:
            if state[neighbor] == color:  # color used in neighbor
                return None

        new_state = state.copy()
        new_state[node] = color
        cost = sum(self.n_nodes ** i for i in range(color + 1))
        return cost, new_state

    def _init_random(self, size, seed):
        """Initializes the problem creating a randomly connected graph."""
        self.n_nodes = size
        self.adjacency = [[False] * self.n_nodes for _ in range(self.n_nodes)]
        random.seed(seed)

        for i in range(self.n_nodes):
            for j in range(i + 1, self.n_nodes):
                adjacent = random.random() < 0.5
                self.adjacency[i][j] = adjacent
                self.adjacency[j][i] = adjacent

    def _init_petersen(self):
        """Initializes the problem using the petersen graph.

        This initialization numbers the nodes clock-wise starting with the
        inner nodes and then the outer ones. This way, the top node of the
        inner star is the 0 node and the top node of the outer pentagon is
        the 5th node.
        """
        self.n_nodes = 10
        self.adjacency = [[False] * self.n_nodes for _ in range(self.n_nodes)]
        edges = [(0, 2), (0, 3), (1, 3), (1, 4), (2, 4),
                 (0, 5), (1, 6), (2, 7), (3, 8), (4, 9),
                 (5, 6), (6, 7), (7, 8), (8, 9), (9, 5)]

        for i, j in edges:
            self.adjacency[i][j] = True
            self.adjacency[j][i] = True


############################### Remedial Exam ################################


class AIExam1617RP2(Problem):
    """Problem 2 [University of Lleida AI Remedial Exam 2015-16]
    Considera el siguiente grafo dirigido del espacio de estados de un problema
    donde el estado A es el estado inicial y G1;G2;G3 son tres estados finales.

    Indica para los siguientes algoritmos los 5 primeros nodos expandidos en
    orden.
        (0.4 puntos) DFS, en grafo, y test de solución en sucesores.
        (0.4 puntos) BFS, en árbol, y test de solución en sucesores.
        (0.4 punto) UCS, en grafo.
        (0.4 puntos) Búsqueda el primero voraz, en árbol.
        (0.4 puntos) A*, en grafo.
        (0.5 puntos) ¿Es la heurística admisible? Razona tu respuesta.
    Notas:
        Los empates se rompen alfabéticamente.
    """

    def __init__(self):
        self.graph = {
            ('A', 'B'): 2, ('A', 'C'): 4, ('A', 'D'): 4,
            ('B', 'A'): 2, ('B', 'C'): 1, ('B', 'E'): 4,
            ('C', 'B'): 1, ('C', 'F'): 3,
            ('D', 'C'): 2, ('D', 'F'): 4, ('D', 'G1'): 12,
            ('E', 'C'): 2, ('E', 'G2'): 4,
            ('F', 'H'): 2, ('F', 'G3'): 6,
            ('H', 'G3'): 1,
        }

        self.heuristic = {
            'A': 9, 'B': 8, 'C': 4, 'D': 6, 'E': 3, 'F': 3, 'H': 2,
            'G1': 0, 'G2': 0, 'G3': 0
        }

        self.nodes = tuple(self.heuristic.keys())
        self.start_state = 'A'
        self.goal_states = frozenset(['G1', 'G2', 'G3'])

    def get_start_states(self):
        """Creates and returns the initial state."""
        return [self.start_state]

    def is_goal_state(self, state):
        """Tests if state is one of the goal states."""
        return state in self.goal_states

    def is_valid_state(self, _):
        """Tests if the state is valid.
        Always evaluates to true because the actions only return valid states.
        """
        return True

    @action(DCategorical('nodes'))
    def move(self, state, n_node):
        """Moves the state to another node.

        The move is valid iff there is an edge connecting the current
        and the next node.
        """
        edge = (state, n_node)

        if edge in self.graph:
            return self.graph[edge], n_node

        return None


################################ Generic code ################################

@AIExam1617P3.heuristic
@AIExam1617RP2.heuristic
class Fixed(Heuristic):
    """Uses a fixed heuristic defined in the problem.

    The instance of the problem class can be accessed via 'self' as
    'self.problem'.
    """

    def compute(self, state):
        """Returns the heuristic value defined in the problem."""
        return self.problem.heuristic[state]
